"""Unique format for all the versions."""

import re
from typing import NamedTuple, Optional

from .error import InvalidVersionFormatError


class Version(NamedTuple):
    """Class containing all the version info."""

    major: int
    minor: Optional[int] = None
    patch: Optional[int] = None
    prerelease: Optional[str] = None
    build: Optional[str] = None


_REGEX = re.compile(
    r"""
        ^
        (?P<major>(?:0|[1-9][0-9]*))
        \.
        (?P<minor>(?:0|[1-9][0-9]*))
        \.
        (?P<patch>(?:0|[1-9][0-9]*))
        (\-(?P<prerelease>
            (?:0|[1-9A-Za-z-][0-9A-Za-z-]*)
            (\.(?:0|[1-9A-Za-z-][0-9A-Za-z-]*))*
        ))?
        (\+(?P<build>
            [0-9A-Za-z-]+
            (\.[0-9A-Za-z-]+)*
        ))?
        $
        """,
    re.VERBOSE,
)


def parse_semver(version_string: str, remove_initial_v=False) -> Version:
    """Parse the semver version.

    credit to https://github.com/k-bx/python-semver

    Args:
        version_string: the string to parse.
        remove_initial_v: strip the initial "v" if the version string has it.
    Returns:
        The semver.

    """
    if remove_initial_v and version_string.startswith(("v", "V")):
        version_string = version_string[1:]

    match = _REGEX.match(version_string)
    if match is None:
        raise InvalidVersionFormatError(
            version_string, "It is not valid SemVer string."
        )

    version_parts = match.groupdict()

    return Version(
        int(version_parts["major"]),
        int(version_parts["minor"]),
        int(version_parts["patch"]),
        version_parts["prerelease"],
        version_parts["build"],
    )



"""Module containing all the errors of the package."""

from typing import Optional


class InvalidVersionFormatError(ValueError):
    """Error thrown when the format is invalid."""

    def __init__(
        self,
        version: str,
        message: Optional[str] = None,
        caused_by: Optional[Exception] = None,
    ):
        """Constructor.

        Args:
            version: The invalid version string.
            message: A explicit reason why the version is invalid.
            caused_by: Underlying error thrown.

        """
        full_message = f"Invalid version {version}."
        if message:
            full_message += "\n" + message
        if caused_by is not None:
            super().__init__(full_message, caused_by)
        else:
            super().__init__(full_message)



"""Module to get the Java version."""

import re
from os import environ, path
from pathlib import Path
from typing import Optional, Union

from universions._exec import exec_command
from universions.error import InvalidVersionFormatError
from universions.version import Version

JAVA_HOME_VAR = "JAVA_HOME"

_JAVA_REGEX = re.compile(
    r"""
        ^
        (?P<major>(?:0|[1-9][0-9]*))
        (
            \.(?P<minor>(?:0|[1-9][0-9]*))
        )?
        (
            \.(?P<patch>(?:0|[1-9][0-9]*))
        )?
        (\-(?P<prerelease>
            (?:0|[1-9A-Za-z-][0-9A-Za-z-]*)
            (\.(?:0|[1-9A-Za-z-][0-9A-Za-z-]*))*
        ))?
        (\_(?P<legacy_patch>
            [0-9]+
        ))?
        $
        """,
    re.VERBOSE,
)


def get_java_version(java_path: Optional[Union[Path, str]] = None) -> Optional[Version]:
    """Get the Java versions.

    Args:
        java_path: The path to the Java to check. If not defined, it uses JAVA_HOME
                  from the enviroment variables to try $JAVA_HOME/bin/java.
                  If JAVA_HOME is not defined either it tries only "java".
    Returns:
        The Java version.

    """
    if java_path is None:
        if JAVA_HOME_VAR in environ:
            java_path = path.join(environ[JAVA_HOME_VAR], "bin", "java")
        else:
            java_path = "java"
    if isinstance(java_path, Path):
        java_path = str(java_path)
    try:
        cmd_result = _get_command_result(java_path)
        version_string = _parse_version_string(cmd_result)
        return _parse_version(version_string)
    except Exception as exe:
        raise Exception("Could not get the Java version.", exe)


def _get_command_result(java_path: str) -> str:
    """Get the result of the command "java -version".

    Args:
        java_path: The path to use for Java in the command.
    Returns:
        The result of the command.

    """
    return exec_command([java_path, "-version"], use_stderr=True)


def _parse_version(version_string: str) -> Version:
    """Parse the version string to return a version.

    Supported versions strings include :
      - 1.8.0_151
      - 11.0.2
      - 14-ea

    Args:
        version_string: The version string such as "1.8.0_151"
    Returns:
        The Java version.
    """
    #  First split the prerelease number if any :
    match = _JAVA_REGEX.match(version_string)
    if match is None:
        raise InvalidVersionFormatError(
            version_string, "It is not valid Java version string."
        )

    parts = match.groupdict()
    major = int(parts["major"])
    minor = int(parts["minor"]) if parts["minor"] is not None else None
    if parts["legacy_patch"] is not None:
        patch = int(parts["legacy_patch"])
    elif parts["patch"] is not None:
        patch = int(parts["patch"])
    else:
        patch = None
    prerelease = parts["prerelease"]
    return Version(major, minor, patch, prerelease, None)


def _parse_version_string(cmd_result: str) -> str:
    """Parse the command result into a string.

    Args:
        cmd_result: The result ofthe command to get the version.
    Results:
        The string of the version such as "1.8.0_151"

    """
    lines = cmd_result.splitlines()
    split_lines = [line.split(" ") for line in lines]
    version_line = [
        line for line in split_lines if len(line) > 0 and line[1] == "version"
    ][0]
    version_string = version_line[2].replace('"', "")
    return version_string



"""Module for node versions."""

import subprocess
from pathlib import Path
from typing import Optional, Union

from universions._exec import exec_command
from universions.version import Version, parse_semver


def get_node_version(node_path: Optional[Union[Path, str]] = None) -> Optional[Version]:
    """Get the Java versions.

    Args:
        node_path: The path to the node to check. If not defined, it uses "node"
    Returns:
        The Node version.

    """
    if node_path is None:
        node_path = "node"
    version_string = exec_command([node_path, "--version"], use_stderr=True)
    return parse_semver(version_string, remove_initial_v=True)



"""Module for Python version.

The version comes from the Python in command line and not the one running the script.

Python version can be get by calling ``python --version`` or ``python -V`

The returned value is "Python" followed by the version number :
    - "Python 2.5.1"
    - "Python 3.6.9"
Possible variations :
    - incomplete number : "Python 3.0"
    - prerelease number :
        - "Python 3.7.5rc1"
        - "Python 3.8.0b2+"
        - "Python 3.9.0a0"
    - Jython : "Jython 2.5.4rc1"
    - Multiline :
        "Python 3.6.9 (5da45ced70e515f94686be0df47c59abd1348ebc, Oct 17 2019, 22:59:56)"
        "[PyPy 7.2.0 with GCC 8.2.0]"

It is not a valid semver number as the prerelease value does not start with "-"

"""

from pathlib import Path
from typing import Optional, Union

from .._exec import exec_command
from ..version import Version
from ._parser import parse_python_version


def get_python_version(path: Optional[Union[Path, str]] = None) -> Optional[Version]:
    """Get the Python version.

    Args:
        path: The path to the Python to check. If not defined, it uses "python"
    Returns:
        The Python version.

    """
    if path is None:
        path = "python"
    version_string = exec_command([path, "--version"], use_stderr=True)
    version = version_string.split(" ")[1]

    return parse_python_version(version)



"""Parser for python version."""

import re

from ..error import InvalidVersionFormatError
from ..version import Version

PYTHON_REGEX = re.compile(
    r"""
        ^
        (?P<major>(?:0|[1-9][0-9]*))
        \.
        (?P<minor>(?:0|[1-9][0-9]*))
        (
            \.
            (?P<patch>(?:0|[1-9][0-9]*))
        )?
        (?P<prerelease>
            (?:[A-Za-z-][0-9A-Za-z-]*\+?)
        )?
        $
        """,
    re.VERBOSE,
)


def parse_python_version(version: str) -> Version:
    """Parse the python version.

    Args:
        version: the version string such as "3.9.0a0"
    Returns:
        THe parsed version object.

    """
    match = PYTHON_REGEX.match(version)
    if match is None:
        raise InvalidVersionFormatError(
            version, "It is not valid Python version string."
        )

    version_parts = match.groupdict()

    return Version(
        int(version_parts["major"]),
        int(version_parts["minor"]),
        int(version_parts["patch"]) if version_parts["patch"] is not None else 0,
        version_parts["prerelease"],
    )



"""Main file to call universions from the CLI."""

import argparse

from universions.java import get_java_version
from universions.node import get_node_version
from universions.python import get_python_version
from universions.version import parse_semver

from ._version import VERSION as universions_version
from .version import Version


def get_self_version() -> Version:
    """Gets the version of this exact module."""
    version = universions_version.replace(".dev", "-dev")
    return parse_semver(version)


TOOLS = {
    "java": get_java_version,
    "node": get_node_version,
    "python": get_python_version,
    "universions": get_self_version,
}


def parse_args():
    """Parses the arguments of the program."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "tool",
        help="Select the tool whose version is wanted",
        choices=list(TOOLS.keys()),
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="Configures the verbosity level of the printed version",
        action="count",
        default=0,
    )
    return parser.parse_args()


def print_version(version, verbosity) -> str:
    """Converts the version into the string matching the verbosity level."""
    if verbosity == 0:
        return f"{version.major}.{version.minor}"
    if 1 <= verbosity < 3:
        return f"{version.major}.{version.minor}.{version.patch}"
    return str(version)


def main() -> int:
    """Main function to be run by the CLI tool."""
    args = parse_args()

    tool_name = args.tool
    get_version = TOOLS.get(tool_name)
    version = get_version()
    print(print_version(version, args.verbosity))

    return 0


if __name__ == "__main__":
    main()



"""Setup universions module."""

from os import path

from setuptools import find_packages, setup


def get_long_description():
    """Read the description from external file."""
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, "README.md"), encoding="utf-8") as file:
        return file.read()


def get_package_version():
    """Read the version of the package.
    See https://packaging.python.org/guides/single-sourcing-package-version
    """
    version_exports = {}
    file_directory = path.abspath(path.dirname(__file__))
    with open(path.join(file_directory, "universions", "_version.py")) as file:
        exec(file.read(), version_exports)  # pylint: disable=exec-used
    return version_exports["VERSION"]


setup(
    name="universions",
    version=get_package_version(),
    author="Fabien Celier",
    author_email="fabien.celier@polytechnique.org",
    packages=find_packages(exclude=["test"]),
    url="https://github.com/fabiencelier/universions",
    keywords=["version", "versions", "universal"],
    license="LICENSE.txt",
    description="Get version of other tools.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    entry_points={"console_scripts": ["universions = universions:cli"]},
)



"""Module to execute a command line and clean it."""

import subprocess
from pathlib import Path
from subprocess import PIPE, Popen
from typing import List, Union


def exec_command(
    parts: List[Union[str, Path]], clean: bool = True, use_stderr: bool = False
) -> str:
    """Exec the command and clean the result.

    Args:
        parts: Part of the command to execute.
        clean: Whether to clean the result.
        use_stderr: Whether to use the stderr
    Returns:
        The result of the command.

    """
    #  cmd = " ".join([str(part) for part in parts])
    cmd = [str(part) for part in parts]
    stderr = subprocess.STDOUT if use_stderr else PIPE
    result_string, _ = Popen(cmd, stdout=PIPE, stderr=stderr).communicate()
    if isinstance(result_string, bytes):
        result_string = str(result_string, encoding="utf-8")
    if clean:
        result_string = result_string.strip()
    return result_string
